import java.util.Vector;
import java.util.Enumeration;

class CandidateRecord {
	// OVERVIEW: A mutable datatype that records the number of votes for a 
	//     candidate.  A typical CandidateRecord is < name, votes > where
	//     name is a String that is the name of the candidate, and votes
	//     is a non-negative number that is the number of votes for the
	//     candidated.

	private String name;
	private int votes;

	// Rep invariant: 
	//     RI (r) = r.votes >= 0 && r.name != null

	// Abstraction function:
	//     AF (r) = < name, votes >

	public CandidateRecord(String n)
	// EFFECTS: Initializes this to < n, 0 >.
	{
		name = n;
		votes = 0;
	}

	public void recordVote()
	// MODIFIES: this
	// EFFECTS: Adds one vote to the vote count for this.  
	//    If this_pre = < n, c > then this_post = < n, c + 1 >.
	{
		votes++;
	}

	public String getName()
	// EFFECTS: Returns the name of this.
	{
		return name;
	}

	public int getVotes()
	// EFFECTS: Returns the number of votes of this.
	{
		return votes;
	}
}

class OfficeResult {
	// OVERVIEW: A mutable datatype that represents the result for one
	//    race.  A typical OfficeResult is
	//       < office, [ < candidate_0, votes_0 >, ..., < candidate_n-1, votes_n-1 > ] >

	// Rep:
	private String office;
	private Vector candidates; // A vector of CandidateRecord objects

	// Rep Invariant
	//   RI(r) =    r.office != null
	//           && r.candidates != null
	//           && r.candidates.elementType == \type(CandidateRecord)
	//           && no duplicate candidate names in r.candidates

	// Abstraction Function
	//    AF(r) = < office, [ <candidate_0, votes_0 >, ..., < candidate_n-1, votes_n-1 > ] >
	//      where candidate_i = r.candidates[i].name
	//            votes_i = r.candidates[i].votes

	public OfficeResult(String o)
	// EFFECTS: Initializes this to < o, [ ] >.
	{
		office = o;
		candidates = new Vector();
	}

	public void addCandidate(String c)
	// REQUIRES: c does not match the name of a candidate in this.
	// EFFECTS: Adds the candidate c to this with no votes.
	//      If this_pre = 
	//           < office, [ < candidate_0, votes_0 >, ..., < candidate_n-1, votes_n-1 > ]
	//      then this_post = 
	//           < office, [ < candidate_0, votes_0 >, ..., < candidate_n-1, votes_n-1 >, < c, 0 > ]
	{
		candidates.add(new CandidateRecord(c));
	}

	private CandidateRecord getRecord(String c)
		throws InvalidCandidateException
	// EFFECTS: If c matches the name of a candidate in this, returns 
	//     the CandidateRecord in this.candidates whos name matches c.
	{
		for (int i = 0; i < candidates.size(); i++) {
			CandidateRecord cr = (CandidateRecord) candidates.get(i);
			if (cr.getName().equals(c)) {
				return cr;
			}
		}

		throw new InvalidCandidateException();
	}

	public String getOffice() {
		// EFFECTS: Returns the office of this.
		return office;
	}

	public void recordVote(String c) throws InvalidCandidateException
	// EFFECTS: If c matches the name of a candidate in this, records
	//     a vote for c.  Otherwise, throws InvalidCandidateException.
	//     If candidate_i = c, this_post.votes_i = this_pre.votes_i + 1.
	{
		try {
			getRecord(c).recordVote();
		} catch (InvalidCandidateException e) {
			throw e;
		}
	}

	public int getVotes(String c) throws InvalidCandidateException
	// EFFECTS: If c matches the name of a candidate in this, returns
	//     the number of votes for c.  Otherwise, throws
	//     InvalidCandidateException.  If candidate_i = c, returns
	//     votes_i.
	{
		try {
			return getRecord(c).getVotes();
		} catch (InvalidCandidateException e) {
			throw e;
		}
	}

	public Vector getCandidates() {
		Vector res = new Vector ();
		for (Enumeration enum = candidates.elements (); enum.hasMoreElements (); ) {
			CandidateRecord r = (CandidateRecord) enum.nextElement ();
			res.add (r.getName());
		}
		return res;
	}
}

public class ElectionResults {
	// OVERVIEW: ElectionResults is a mutable datatype for recording the
	//    results of an election.  A typical ElectionResults is:
	//
	//        [ < office_0, results_0 >, < office_1, results_1 >, ..., < office_n-1, results_n-1 > ]
	//
	//    where office_i is the name of the ith office on the ballot and
	//
	//        results_i = [ < candidate_i_0, votes_i_0 > , ..., < candidate_i_m-1, votes_i_m-1 > ]
	//
	//     is a list of the candidates for office_i and the number of
	//     votes they received.
	//
	//     For example, 
	//
	//        [ < "Sheriff", [ < "Officer Krispy", 2043 >, < "Alice Enforcer", 3024 > ] >,
	//          < "Dog Catcher", [ < "Spot", 23 >, < "Muffy", 12 > ] > ]
	//      
	//     describes an election with two offices.  
	//

	// Rep:
	private Vector results; // A vector of OfficeResult objects

	// Rep Invariant:
	//     < missing >
	//
	// Abstraction Function:
	//     < missing >

	public ElectionResults()
	// EFFECTS: Initializes this to an election with no offices.
	{
		results = new Vector();
	}

	public ElectionResults(BallotDefinition bd)
	// EFFECTS: Initializes this to an election with all offices and
	//   candidates in bd with 0 votes.  For texample, if 
	//      bd = [ <"Sheriff", [ "Officer Krispy", "Alice Enforcer" ] >,
	//             <"Dog Catcher", [ "Spot", "Muffy" ] > ]
	//    then,
	//      this_post = 
	//        [ < "Sheriff", [ < "Officer Krispy", 0 >, < "Alice Enforcer", 0 > ] >,
	//          < "Dog Catcher", [ < "Spot", 0 >, < "Muffy", 0 > ] > ]
	{
		// ... // implementation not shown
	}

	public void addOffice(Office o)
	// REQUIRES: this does not contain an office with the same name
	//              as o.
	// EFFECTS: Adds the office o to this election.
	{
		OfficeResult or = new OfficeResult(o.getName());
		for (Enumeration candidates = o.candidates();
			candidates.hasMoreElements();
			) {
			or.addCandidate((String) candidates.nextElement());
		}

		results.add(or);
	}

	private OfficeResult lookupOffice(String s) throws NoSuchOfficeException
	// EFFECTS: If s is the name of an office in this election,
	//             returns the OfficeResult object associated with s.
	//             Otherwise, throws NoSuchOfficeException.

	{
		for (Enumeration offices = results.elements();
			offices.hasMoreElements();
			) {
			OfficeResult or = (OfficeResult) offices.nextElement();
			if (or.getOffice().equals(s)) {
				return or;
			}
		}

		throw new NoSuchOfficeException();
	}

	public void recordVote(String office, String candidate)
		throws NoSuchOfficeException, InvalidCandidateException
	// MODIFIES: this
	// EFFECTS: If office does not match the name of an office in this,
	//    throws InvalidOfficeException.  If candidate does not match
	//    the name of a candidate for office, throws
	//    InvalidCandidataException.  Otherwise, increments the number
	//    of votes for candidate by 1.
	// 
	//    For example, if this_pre = 
	//        [ < "Sheriff", [ < "Officer Krispy", 0 >, < "Alice Enforcer", 0 > ] >,
	//          < "Dog Catcher", [ < "Spot", 3 >, < "Muffy", 4 > ] > ]
	//     and office = "Dog Catcher" and candidate = "Muffy", then
	//     this_post = 
	//        [ < "Sheriff", [ < "Officer Krispy", 0 >, < "Alice Enforcer", 0 > ] >,
	//          < "Dog Catcher", [ < "Spot", 3 >, < "Muffy", 5 > ] > ]
	{
		try {
			OfficeResult ofr = lookupOffice(office);
			ofr.recordVote(candidate);
		} catch (NoSuchOfficeException e) {
			throw e;
		} catch (InvalidCandidateException e) {
			throw e;
		}
	}

	public int getVotes(String office, String candidate)
		throws NoSuchOfficeException, InvalidCandidateException {
		try {
			OfficeResult ofr = lookupOffice(office);
			return ofr.getVotes(candidate);
		} catch (NoSuchOfficeException e) {
			throw e;
		} catch (InvalidCandidateException e) {
			throw e;
		}
	}

	public String getWinner(String office)
		throws NoSuchOfficeException, NoWinnerException
	// EFFECTS: If office does not match the name of an office in this,
	//    throws NoSuchOfficeException.  If there are no candidates for
	//    office with any votes, throws NoWinnerException.  If there are two
	//    or more candidates for office with the same number of votes, and no
	//    candidates for office with more votes, throws
	//    NoWinnerException.  Otherwise, return the name of the candidate
	//    for office who received more votes than every other candidate.
	//    For example, if this_pre =
	//        [ < "Sheriff", [ < "Officer Krispy", 3 >, < "Alice Enforcer", 12 > ] >,
	//          < "Dog Catcher", [ < "Spot", 4 >, < "Muffy", 4 > ] > ]
	//    getWinner ("Sheriff") returns "Alice Enforcer" and
	//    getWinner ("Dog Catcher") throws NoWinnerException.
	{
		OfficeResult or = lookupOffice(office);
		// note this will re-throw the NoSuchOfficeException
		String winner = null;
		int maxvotes = 0;

		Vector candidates = or.getCandidates();
		for (Enumeration candenum = candidates.elements();
			candenum.hasMoreElements();
			) {
			String candidate = (String) candenum.nextElement();
			try {
				int cvotes = getVotes(office, candidate);

				if (cvotes == maxvotes) {
					winner = null;
				} else if (cvotes > maxvotes) {
					maxvotes = cvotes;
					winner = candidate;
				}
			} catch (InvalidCandidateException ice) {
				throw new RuntimeException("BUG: should not have invalid candidates here!");
				}
		}

		if (winner != null) {
			return winner;
		} else {
			throw new NoWinnerException();
		}
	}
}
